import os
import shutil
from contextlib import suppress
from dataclasses import dataclass
from pathlib import Path

from pytest_results._core.storages.abc import Storage


@dataclass(repr=False, eq=False, frozen=True, slots=True)
class LocalStorage(Storage):
    dir: Path
    temporary_dir: Path

    def copy(self, filepath: Path, destination: Path) -> None:
        shutil.copyfile(filepath, destination)

    def exists(self, path: Path) -> bool:
        return path.exists()

    def get_absolute_path(self, relative_path: Path) -> Path:
        return self.__get_absolute_path(relative_path, self.dir)

    def get_temporary_path(self, relative_path: Path) -> Path:
        return self.__get_absolute_path(relative_path, self.temporary_dir)

    def read(self, filepath: Path) -> bytes:
        with suppress(FileNotFoundError):
            with open(filepath, "rb") as reader:
                return reader.read()

        return b""

    def write(self, filepath: Path, result: bytes = b"") -> None:
        os.makedirs(filepath.parent, exist_ok=True)

        with open(filepath, "wb") as writer:
            writer.write(result)

    @staticmethod
    def __get_absolute_path(filepath: Path, directory: Path) -> Path:
        assert not filepath.is_absolute()
        return directory / filepath



from contextlib import suppress
from functools import singledispatch
from typing import Any

import orjson


@singledispatch
def json_dump(value: Any) -> bytes:
    return orjson.dumps(value, default=str, option=orjson.OPT_INDENT_2)


@json_dump.register
def _(value: bytes) -> bytes:
    return value


with suppress(ImportError):
    import msgspec

    @json_dump.register
    def _(value: msgspec.Struct) -> bytes:
        b = msgspec.json.encode(value)
        return msgspec.json.format(b, indent=2)


with suppress(ImportError):
    import pydantic

    @json_dump.register
    def _(value: pydantic.BaseModel) -> bytes:
        return value.model_dump_json(indent=2).encode()



from abc import abstractmethod
from collections.abc import Callable, Sequence
from dataclasses import dataclass
from pathlib import Path
from types import TracebackType
from typing import Any, Final, Protocol, Self, runtime_checkable

from pytest_results._core.dump_functions.json import json_dump
from pytest_results._core.storages.abc import Storage
from pytest_results.exceptions import ResultsMismatchError

type DumpFunction[T] = Callable[[T], bytes]
type CommandRunner = Callable[[str], Any]

_DEFAULT_DUMP_FUNCTION: Final[DumpFunction[Any]] = json_dump
_DEFAULT_FILE_FORMAT: Final[str] = "json"


@runtime_checkable
class Regression(Protocol):
    __slots__ = ()

    @abstractmethod
    def check[T](
        self,
        current_result: T,
        /,
        suffix: str | None = ...,
        dump_func: DumpFunction[T] | None = ...,
        file_format: str | None = ...,
    ) -> None:
        raise NotImplementedError


@dataclass(repr=False, eq=False, frozen=True, slots=True)
class BoundRegression(Regression):
    regression: Regression
    dump_func: DumpFunction[Any]
    file_format: str

    def check[T](
        self,
        current_result: T,
        /,
        suffix: str | None = None,
        dump_func: DumpFunction[T] | None = None,
        file_format: str | None = None,
    ) -> None:
        __tracebackhide__ = True
        return self.regression.check(
            current_result,
            suffix,
            dump_func or self.dump_func,
            file_format or self.file_format,
        )


@dataclass(repr=False, eq=False, frozen=True, slots=True)
class RegressionImpl(Regression):
    testinfo: Sequence[str]
    storage: Storage
    command_runner: CommandRunner

    def check[T](
        self,
        current_result: T,
        /,
        suffix: str | None = None,
        dump_func: DumpFunction[T] | None = None,
        file_format: str | None = None,
    ) -> None:
        __tracebackhide__ = True

        dump_func = dump_func or _DEFAULT_DUMP_FUNCTION
        file_format = file_format or _DEFAULT_FILE_FORMAT
        suffix = suffix or ""

        current_bytes = dump_func(current_result)
        relative_filepath = self.__get_relative_result_filepath(file_format, suffix)
        filepath = self.storage.get_absolute_path(relative_filepath)
        previous_bytes = self.storage.read(filepath)

        try:
            assert current_bytes == previous_bytes

        except AssertionError as exc:
            temporary_filepath = self.storage.get_temporary_path(relative_filepath)
            self.storage.write(temporary_filepath, current_bytes)
            raise ResultsMismatchError(
                current=temporary_filepath,
                previous=filepath,
                storage=self.storage,
                command_runner=self.command_runner,
            ) from exc

    def __get_relative_result_filepath(self, file_format: str, suffix: str) -> Path:
        testinfo = self.testinfo
        filename = f"{testinfo[-1]}{suffix}.{file_format}"
        return Path(*testinfo[:-1], filename)


class RegressionStack(Regression):
    __slots__ = ("__count", "__delegate", "__mismatches")

    __count: int
    __delegate: Regression
    __mismatches: list[ResultsMismatchError]

    def __init__(self, delegate: Regression) -> None:
        self.__count = 0
        self.__delegate = delegate
        self.__mismatches = []

    def __enter__(self) -> Self:
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> None:
        __tracebackhide__ = True

        if mismatches := self.__mismatches:
            raise (
                mismatches[0]
                if len(mismatches) == 1
                else ExceptionGroup("", mismatches)
            )

    def check[T](
        self,
        current_result: T,
        /,
        suffix: str | None = None,
        dump_func: DumpFunction[T] | None = None,
        file_format: str | None = None,
    ) -> None:
        __tracebackhide__ = True

        if not suffix and (count := self.__count) > 0:
            suffix = f"_{count}"

        self.__count += 1

        try:
            return self.__delegate.check(current_result, suffix, dump_func, file_format)
        except ResultsMismatchError as mismatch:
            self.__mismatches.append(mismatch)



import os
import warnings
from collections.abc import Awaitable, Callable, Generator, Iterable, Mapping, Sequence
from functools import update_wrapper
from inspect import iscoroutinefunction
from pathlib import Path
from typing import Any, ClassVar

import pytest

from pytest_results import _LocalStorage, _RegressionImpl, _RegressionStack
from pytest_results._exc_group import iter_nested_exceptions
from pytest_results._testinfo import get_testinfo
from pytest_results.exceptions import ResultsMismatchError

__all__ = ()


class PytestResultsConfig:
    __slots__ = ("__config",)

    __diff_commands: ClassVar[Mapping[str, str]] = {
        "cursor": "cursor -d -r -w {current} {previous}",
        "pycharm": "pycharm diff {current} {previous}",
        "vscode": "code -d -r -w {current} {previous}",
    }

    def __init__(self, pytest_config: pytest.Config) -> None:
        self.__config = pytest_config

    @property
    def accept_diff(self) -> bool:
        return self.__config.getoption("accept_diff")

    @property
    def diff_command(self) -> str | None:
        if diff := self.__get_option_or_ini("diff"):
            return diff

        if ide := self.__get_option_or_ini("ide"):
            lowercase_ide = ide.lower()

            try:
                return self.__diff_commands[lowercase_ide]
            except KeyError:
                warnings.warn(f"pytest-results doesn't yet support the `{ide}` IDE.")

        return None

    def __get_option_or_ini[T](self, key: str) -> T | None:
        config = self.__config
        return config.getoption(key, default=config.getini(key))


@pytest.hookimpl
def pytest_addoption(parser: pytest.Parser) -> None:
    group = parser.getgroup("pytest-results")

    group.addoption(
        "--accept-diff",
        dest="accept_diff",
        action="store_true",
        help="Parameter for accepting new differences between results.",
        default=False,
    )

    diff_help = "Command line to open an interactive comparison. Example: `code -d -w {current} {previous}`."
    group.addoption(
        "--diff",
        dest="diff",
        metavar="COMMAND_LINE",
        help=diff_help,
        default=None,
    )
    parser.addini(
        "diff",
        type="string",
        help=diff_help,
        default=None,
    )

    ide_help = "The IDE to open for interactive comparison."
    group.addoption(
        "--ide",
        dest="ide",
        metavar="IDE",
        help=ide_help,
        default=None,
    )
    parser.addini(
        "ide",
        type="string",
        help=ide_help,
        default=None,
    )


@pytest.hookimpl
def pytest_collection_modifyitems(items: Iterable[pytest.Item]) -> None:
    for item in items:
        if isinstance(item, pytest.Function):
            __autodetect_result(item)


@pytest.hookimpl(trylast=True, wrapper=True)
def pytest_pyfunc_call(
    pyfuncitem: pytest.Function,
) -> Generator[None, object | None, object | None]:
    __tracebackhide__ = True

    try:
        result = yield

    except ResultsMismatchError as mismatch:
        __on_mismatches((mismatch,), pyfuncitem.config)
        raise mismatch

    except ExceptionGroup as exc_group:
        if sub_exc_group := exc_group.subgroup(ResultsMismatchError):
            mismatches = tuple(iter_nested_exceptions(sub_exc_group))
            __on_mismatches(mismatches, pyfuncitem.config)

        raise exc_group

    return result


@pytest.fixture(scope="function")
def regression(request: pytest.FixtureRequest, tmp_path: Path) -> _RegressionStack:
    dirname = "__pytest_results__"
    storage = _LocalStorage(request.config.rootpath / dirname, tmp_path / dirname)
    testinfo = get_testinfo(request)
    delegate = _RegressionImpl(testinfo, storage, os.system)
    return _RegressionStack(delegate)


def __autodetect_result(pyfuncitem: pytest.Function) -> pytest.Function:
    wrapper: Callable[..., None | Awaitable[None]]
    wrapped = pyfuncitem.obj

    if iscoroutinefunction(wrapped):

        async def wrapper(*args: Any, **kwargs: Any) -> None:
            with __get_regression_fixture(pyfuncitem) as regression:
                if (result := await wrapped(*args, **kwargs)) is not None:
                    regression.check(result)

    else:

        def wrapper(*args: Any, **kwargs: Any) -> None:
            with __get_regression_fixture(pyfuncitem) as regression:
                if (result := wrapped(*args, **kwargs)) is not None:
                    regression.check(result)

    pyfuncitem.obj = update_wrapper(wrapper, wrapped)
    return pyfuncitem


def __get_regression_fixture(pyfuncitem: pytest.Function) -> _RegressionStack:
    return pyfuncitem._request.getfixturevalue(regression.__name__)


def __on_mismatches(
    mismatches: Sequence[ResultsMismatchError],
    pytest_config: pytest.Config,
) -> None:
    if not mismatches:
        return

    config = PytestResultsConfig(pytest_config)

    if config.accept_diff:
        for mismatch in mismatches:
            mismatch.accept_diff()

        pytest.skip()

    elif command := config.diff_command:
        for mismatch in mismatches:
            mismatch.show_diff(command)



from collections.abc import Iterator

import pytest

__all__ = ("get_testinfo",)


def get_testinfo(request: pytest.FixtureRequest) -> tuple[str, ...]:
    return tuple(__iter_testinfo(request))


def __iter_testinfo(request: pytest.FixtureRequest) -> Iterator[str]:
    yield from request.module.__name__.split(".")

    if cls := request.cls:
        yield cls.__name__

    yield request.function.__name__



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

if TYPE_CHECKING:  # pragma: no cover
    from pytest_results import _CommandRunner, _Storage

__all__ = ("PytestResultsError", "ResultsMismatchError")


class PytestResultsError(Exception): ...


class ResultsMismatchError(AssertionError, PytestResultsError):
    __slots__ = (
        "__command_runner",
        "__current_filepath",
        "__previous_filepath",
        "__storage",
    )

    __command_runner: _CommandRunner
    __current_filepath: Path
    __previous_filepath: Path
    __storage: _Storage

    def __init__(
        self,
        current: Path,
        previous: Path,
        storage: _Storage,
        command_runner: _CommandRunner,
    ) -> None:
        self.__command_runner = command_runner
        self.__current_filepath = current
        self.__previous_filepath = previous
        self.__storage = storage

    def __str__(self) -> str:
        return self.__to_string(self.current, self.previous)

    @property
    def current(self) -> Path:
        return self.__current_filepath

    @property
    def previous(self) -> Path:
        return self.__previous_filepath

    def accept_diff(self) -> None:
        self.__ensure_file_exists(self.previous)
        self.__storage.copy(self.current, self.previous)

    def show_diff(self, command: str) -> None:
        self.__ensure_file_exists(self.previous)
        command = command.format(current=self.current, previous=self.previous)
        self.__command_runner(command)

    def __ensure_file_exists(self, filepath: Path) -> None:
        if self.__storage.exists(filepath):
            return

        self.__storage.write(filepath)

    @staticmethod
    def __to_string(current: object, previous: object) -> str:
        return f"Results mismatch\n・Current: `{current}`\n・Previous: `{previous}`"



from collections.abc import Iterator


def iter_nested_exceptions[T: Exception](
    exception_group: ExceptionGroup[T],
) -> Iterator[T]:
    for exception in exception_group.exceptions:
        if isinstance(exception, ExceptionGroup):
            yield from iter_nested_exceptions(exception)
            continue

        yield exception



from abc import abstractmethod
from pathlib import Path
from typing import Protocol, runtime_checkable


@runtime_checkable
class Storage(Protocol):
    __slots__ = ()

    @abstractmethod
    def copy(self, filepath: Path, destination: Path) -> None:
        raise NotImplementedError

    @abstractmethod
    def exists(self, path: Path) -> bool:
        raise NotImplementedError

    @abstractmethod
    def get_absolute_path(self, relative_path: Path) -> Path:
        raise NotImplementedError

    @abstractmethod
    def get_temporary_path(self, relative_path: Path) -> Path:
        raise NotImplementedError

    @abstractmethod
    def read(self, filepath: Path) -> bytes:
        raise NotImplementedError

    @abstractmethod
    def write(self, filepath: Path, result: bytes = ...) -> None:
        raise NotImplementedError



from ._core.regression import BoundRegression, Regression
from ._core.regression import CommandRunner as _CommandRunner
from ._core.regression import RegressionImpl as _RegressionImpl
from ._core.regression import RegressionStack as _RegressionStack
from ._core.storages.abc import Storage as _Storage
from ._core.storages.local import LocalStorage as _LocalStorage

__all__ = (
    "BoundRegression",
    "Regression",
    "_CommandRunner",
    "_LocalStorage",
    "_RegressionImpl",
    "_RegressionStack",
    "_Storage",
)
